/*
 *	Simple mail user interface for KA9Q IP/TCP package.
 *	A.D. Barksdale Garbee II, aka Bdale, N3EUA
 *	Copyright 1986 Bdale Garbee, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *	Copyright 1987 1988 Dave Trulli NN2Z, All Rights Reserved.
 *	Permission granted for non-commercial copying and use, provided
 *	this notice is retained.
 *
 *	Ported to NOS at 900120 by Anders Klemets SM0RGV.
 *
 *  Userlogging, 'RM' and 'KM' implementation,
 *  more-prompts for all types
 *  920307 and later, by Johan. K. Reinalda, WG7J/PA3DIS
 *  Mail indexing by Johan. K. Reinalda, WG7J, June/July 1993
 */
#include <ctype.h>
#include <time.h>
#include <conio.h>
#include "global.h"
#include "ftpserv.h"
#include "smtp.h"
#include "proc.h"
#include "usock.h"
#include "socket.h"
#include "telnet.h"
#include "timer.h"
#include "session.h"
#include "files.h"
#include "mailbox.h"
#include "cmdparse.h"
#include "bm.h"
#include "mailutil.h"
#include "dirutil.h"

#define		SETVBUF
#if	defined(UNIX) || defined(MICROSOFT)
#include	<sys/types.h>
#endif
/*
#if	defined(UNIX) || defined(MICROSOFT) || defined(__TURBOC__)
#include	<sys/stat.h>
#endif
#ifdef AZTEC
#include <stat.h>
#endif
*/
#include <fcntl.h>
#include "bm.h"
#include "mailbox.h"

#ifdef SETVBUF
#define		MYBUF	1024
#endif

#ifdef MAILCMDS

char Badmsg[] = "Invalid Message number %d\n";
char Nomail[] = "No messages\n";
static char Noaccess[] = "Unable to access %s\n";

static int readnotes __ARGS((struct mbx *m,FILE *ifile));
static long isnewmail __ARGS((struct mbx *m));
static int initnotes __ARGS((struct mbx *m));
static int lockit __ARGS((struct mbx *m));

static int
initnotes(m)
struct mbx *m;
{
    FILE *idx;
    char *area;
    struct let *cmsg;
    int ret;
    char buf[128];

    /* Reset all mailbox status variables */
    m->change = m->anyread = m->nmsgs = m->current = \
        m->newmsgs = m->mboxsize = 0;
    area = strdup(m->area);
    dirformat(area);
    sprintf(buf,"%s/%s.idx",Mailspool,area);
    if ((idx = fopen(buf,"r")) == NULLFILE)
        /* No mail ! */
        return 0;
    ret = readnotes(m,idx);
    fclose(idx);
    if (ret != 0)
		return -1;

	return 0;
}

/* readnotes assumes that idx is pointing to the index file */
static int
readnotes(m,idx)
struct mbx *m;
FILE *idx;
{
    char buf[LINELEN];
    struct let *oldmbox;
    long pos;
    char status;
    struct let *cmsg;

    if(idx == NULL)
        return 0;

    /* See if the index file is intact - WG7J */
    fseek(idx,0L,SEEK_END);
    pos = ftell(idx);
    if(pos % IDXLINESIZE) {
        mail_error("Index '%s.idx' corrupt!\nRun MAIL2IDX !",m->area);
        return 0;
    }
    pos /= IDXLINESIZE;
    rewind(idx);

    if(m->mbox)
        oldmbox = m->mbox;
    else
        oldmbox = NULL;

    /* Get the new number of messages */
    m->mbox = (struct let *)callocw(pos+1,sizeof(struct let));

    if(oldmbox) {    /* Copy the status of the messages back */
        memcpy(m->mbox,oldmbox,(m->nmsgs+1)*sizeof(struct let));
        free(oldmbox);
    }

    cmsg = m->mbox;
    pos = 0;
    m->nmsgs = 0;
    /* Check for unread(ie. new) messages */
    m->newmsgs = m->current = 0;
    while(fgets(buf,LINELEN,idx) != NULLCHAR) {
        cmsg++;
        m->nmsgs++;
#ifdef USERLOG
        cmsg->msgid = atol(buf);
#endif
        cmsg->size = atol(&buf[40]);
        cmsg->start = pos;
        pos += cmsg->size;
        status = buf[8];
        if(status == 'R')
            cmsg->status |= BM_READ;
        else if (status == 'H')
            cmsg->status |= BM_HOLD;
#ifdef USERLOG
        if(m->areatype == AREA)
            if(cmsg->msgid <= m->lastread)
                cmsg->status |= BM_READ;
#endif
        if(!(cmsg->status & BM_READ)) {
            m->newmsgs++;
            if(m->current == 0)
                m->current = m->nmsgs; /* first new message */
        }
	}
     /* start at one if no new messages */
    if(m->current == 0)
        m->current++;
    m->mboxsize = pos;
    if(m->areatype == PRIVATE)  /* our private mail area */
        m->mycnt = m->nmsgs;

	return 0;
}

static char ListHeader[] = "St.  #  TO            FROM     DATE     SIZE SUBJECT\n";

/* list headers of a notesfile a message */
/* Rearranged display - WG7J */
int
dolistnotes(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
    struct let *cmsg;
    char *cp, *s;
    FILE *idx;
    char *area;
    char type;
    char status;
    int start, stop;
    long pos;
    int i,header,list,listed;
#ifdef USERLOG
	long msgid;
#endif
    long size;
    int c,usemore=0,lin;
    char string[IDXSUBLEN];
    char buf[128];

	m = (struct mbx *) p;

    if(m->nmsgs == 0) {
        tputs(Nomail);
        return 0;
    }

    /* If this user doesn't have read-permissions,
     * we're not going to let him list anything - WG7J
     */
    if(m->privs & NO_READCMD) {
        tputs(Noperm);
        return 0;
    }

    if((m->stype == '>' || m->stype == '<' || m->stype ==  'S') && argc == 1) {
        tputs("Search criterium needed!\n");
        return 0;
    }

    if((lin=m->morerows) != 0)
        usemore = 1;    /* Display More prompt */

    area = strdup(m->area);
    dirformat(area);
    sprintf(buf,"%s/%s.idx",Mailspool,area);
    free(area);
    if((idx = fopen(buf,"r")) == NULL) {
        closenotes(m);
        tprintf("Can not read index file %s\n",buf);
        return 0;
    }

    tprintf("Mail area: %s\n"
            "%d message%s -  %d new\n\n",
            m->area,m->nmsgs,m->nmsgs == 1 ? " " : "s ", m->newmsgs);

    stop = m->nmsgs;
#ifdef USERLOG
    if(m->areatype != PRIVATE && m->stype != 'A')
        start = stop - m->newmsgs + 1;
    else
#endif
    if((argc == 1) & (m->stype == 'L'))
       start = stop;
    else
        start = 1;
    if(argc > 1) {
        if(m->stype == 'L')                     /* LL (List Last) command */
            start = stop - atoi(argv[1]) + 1;
        else
            start = atoi(argv[1]);
        if(argc > 2)
            stop = atoi(argv[2]);

        /* Check the boundaries */
        if(stop > m->nmsgs)
            stop = m->nmsgs;
        if(start < 1)
            start = 1;
    }
    if(start > stop) {
        if((m->stype == ' ') || (m->stype == 'M'))
            tputs("None to list.\n");
        else
            tputs("Invalid range.\n");
        fclose(idx);
        return 0;
	}

    cmsg = m->mbox;
    i = 0;
    header = 0;
    listed = 0;
    while(fgets(buf,sizeof(buf),idx) != NULLCHAR) {
        cmsg++;
        i++;
        if(start <= i && i <= stop) {
            list = 1;
            /* Now see if we should list this one */
            if(m->stype == ' ') {
                if(argc > 1)
                    list = 1;
                else if(cmsg->status & BM_READ)
                    list = 0;
            } else if(m->stype == '>') {
                buf[23] = '\0';
                strcpy(string,&buf[10]);
                if(strstr(strlwr(string),argv[1]) == NULL)
                    list  = 0;
                buf[23] = ' ';
            } else if( m->stype == '<' ||
                      (m->stype == 'M' && m->areatype != PRIVATE)) {
                buf[32] = '\0';
                strcpy(string,&buf[24]);
                if(strstr(strlwr(string),argv[1]) == NULL)
                    list  = 0;
                buf[32] = ' ';
            } else if(m->stype == 'S') {
                strcpy(string,&buf[47]);
                if(strstr(strlwr(string),argv[1]) == NULL)
                    list  = 0;
            } else if(m->stype == 'B' || m->stype == 'T')
                if(m->stype != buf[6])
                    list  = 0;
            if(list) {
#ifdef USERLOG
                /* Check the ID number of this message and
                 * adjust new lastread count, if needed - WG7J
                 */
                if(cmsg->msgid > m->lastread)
                    m->newlastread = cmsg->msgid;
#endif
                if(!header) {
                    tputs(ListHeader);
                    lin--;
                    header = 1;
                }
                listed = 1;
                tputc((i == m->current) ? '>' : ' ');
                tputc(cmsg->status & BM_DELETE ? 'D' : ' ');
                tputc(cmsg->status & BM_HOLD ? 'H':(cmsg->status & BM_READ ? 'Y' : 'N'));
                buf[82] = '\0';   /* Get rid of cr/lf */
                tprintf(" %3d %s",i,&buf[10]);
                lin--;

                /* More prompting added - WG7J */
                if(usemore && lin <= 0){
                    if(m->type == TELNET_LINK || m->type == TIP_LINK)
                        c = tkeywait("--More--",0);
                    else  /* For AX.25 and NET/ROM connects - WG7J */
                        c = mykeywait("More(N=no)? ",m);
                    if(c == -1 || c == 'q' || c == 'Q')
                        break;
                    if(c == '\n' || c == '\r')
                        lin = 1;
                    else
                        lin = m->morerows-1;

                    header = 0;
                }
            }
        }
    }
    fclose(idx);
    if(!listed)
        tputs(Nomail);

	return 0;
}

/*  save msg on stream - if noheader set don't output the header */
int
msgtofile(m)
struct mbx *m;
{
    FILE *fp;
    char *area;
    char buf[LINELEN];

    if (m->nmsgs == 0) {
        tputs(Nomail);
        return -1;
	}

    /* Open the mailbox and go to start of message */
    area = strdup(m->area);
    dirformat(area);
    sprintf(buf,"%s/%s.txt",Mailspool,area);
    if ((fp=fopen(buf,"rt")) == NULL) {
        free(area);
        return -1;
    }
    free(area);
    fseek(fp,m->mbox[m->current].start,0);

    /* skip header */
    while(fgets(buf,sizeof(buf),fp) != NULLCHAR)
        if (*buf == '\n')
            break;

    /* Copy the body */
    while (fgets(buf,sizeof(buf),fp) != NULLCHAR) {
        if(!strncmp(buf,"From ",5))
            break;
        fputs(buf,m->tfile);
        if (ferror(m->tfile)) {
            tputs("Error writing mail file\n");
            fclose(fp);
            return -1;
		}
	}
    fclose(fp);
    return 0;
}

/* dodelmsg - delete message in current notesfile */
/* Modified to allow the 'KM' command. 920307 -  WG7J */
int
dodelmsg(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    struct mbx *m;
	int msg,i;
    char *myargv[NARG];
    int myargc;
    int maxmsg;
    struct let *cmsg;
    char *tmpbuf;
    int start;
    int end;

    m = (struct mbx *) p;

    if (m->nmsgs == 0) {
        tputs(Nomail);
        return 0;
    }
    /* If this user doesn't have read-permissions,
     * we're not going to let him kill anything;
     * allow anyone to kill messages in areas
     * who's names start with 'nts' - WG7J
     */
    /* Check if we have permission to delete others mail */
    if( (m->privs & NO_SENDCMD) ||
        (m->privs & NO_READCMD) ||
        ( !(m->privs & FTP_WRITE) &&
          !(m->areatype == PRIVATE) &&
          strnicmp(m->area,"nts",3)) ){
        tputs(Noperm);
        return 0;
    }
    /* If this is the KM command, setup myargv[]
     * to contain up to NARG message numbers - WG7J
     */
    if(m->stype == 'M') {
        /* scan all messsages to find unread ones */
        cmsg = &m->mbox[1];
        myargc = 1;
        i = 0;
        while(myargc < NARG && i < m->nmsgs) {
            if(cmsg->status & BM_READ) { /*found a read msg!*/
                tmpbuf = mallocw(17); /*allocate space for the new argument*/
                myargv[myargc++] = itoa(i+1,tmpbuf,10);
            }
            i++;
            cmsg++;
        }
        if(myargc == 1) {
            tputs("Nothing to kill\n");
            return 0;
        }
        argc = myargc;
    } else {
        if(argc == 1) {
            msg = m->current;
            if(m->stype == 'U')
              m->mbox[msg].status &= ~BM_DELETE;
            else {
              m->mbox[msg].status |= BM_DELETE;
              m->change = 1;
            }
            tprintf("Msg %d %sKilled.\n", msg,(m->stype=='U') ? "Un-" : "");
            return 0;
        }
        /*simply point to the old arguments*/
        for(i=1;i<argc;i++)
            myargv[i] = argv[i];
    }
    /* See if x - y format was used and use a for i = x to y loop */
    if(myargv[2][0] == '-') {
        start = atoi(myargv[1]);
        end   = atoi(myargv[3]);
        if(start < 0 || start > m->nmsgs || end < 0 || end > m->nmsgs ) {
            tprintf(Badmsg,start);
            start = end-1;
        }
        for(i=start;i<=end;i++) {
            msg = i;
            if(m->stype == 'U')
              m->mbox[msg].status &= ~BM_DELETE;
            else {
              m->mbox[msg].status |= BM_DELETE;
            m->change = 1;
            }
            tprintf("Msg %d %sKilled.\n", msg,(m->stype=='U') ? "Un-" : "");
        } /* endfor */
    } else for(i = 1; i < argc; ++i) {
        tmpbuf = strchr(myargv[i],'-'); /* N5KNX: allow from-to msg specification */
        msg = atoi(myargv[i]);
        if (tmpbuf == NULLCHAR)
            maxmsg = msg;
        else
            maxmsg = atoi(++tmpbuf);
        if (maxmsg < msg) {
            tprintf(Badmsg,myargv[i]);
            continue;
        }
        for (; msg <= maxmsg; msg++) {
            if(msg < 0 || msg > m->nmsgs) {
                tprintf(Badmsg,msg);
                continue;
            }
            if(m->stype == 'U')
                m->mbox[msg].status &= ~BM_DELETE;
            else {
                m->mbox[msg].status |= BM_DELETE;
                m->change = 1;
            }
            tprintf("Msg %d %sKilled.\n", msg,(m->stype=='U') ? "Un-" : "");
        }
    }
    /* If this was 'KM'
     * free the memory allocated for myargv[] - WG7J
     */
    if(m->stype == 'M') {
        for(i=1;i<argc;i++)
            free(myargv[i]);
    }
    return 0;
}

void no_area(struct mbx *m) {

    m->area[0] = '\0';
    m->nmsgs = m->current = m->newmsgs = 0;
    m->mboxsize = 0;
    free(m->mbox);
    m->mbox = NULL;
}

/* close the temp file while copying mail back to the mailbox */
int
closenotes(m)
struct mbx *m;
{
    struct let *cmsg;
    char *area;
    int i,keep,error;
    long start,pos,size;
    FILE *idx,*txt,*new,*idxnew;
    char idxline[IDXLINESIZE+1];
    char len[10],lens[7];
    char buf[LINELEN];
    char old[LINELEN];

    if(m->nmsgs == 0)
        return 0;

    if(!m->change || (isnewmail(m) < 0L)) {
        /* no changes were made,
         * or some messages where deleted !
         * In the latter we cannot continue !
         */
        no_area(m);
		return 0;
	}
    scanmail(m);
	if(lockit(m))
		return -1;
    area = strdup(m->area);
    dirformat(area);
    sprintf(buf,"%s/%s.txt",Mailspool,area);
    if((txt = fopen(buf,"rb")) == NULLFILE) {
        no_area(m);
        tprintf(Noaccess,buf);
		rmlock(Mailspool,m->area);
		return -1;
	}
    sprintf(buf,"%s/%s.idx",Mailspool,area);
    if((idx = fopen(buf,"rt")) == NULLFILE) {
        fclose(txt);
        no_area(m);
		tprintf(Noaccess,buf);
		rmlock(Mailspool,m->area);
		return -1;
    }
    sprintf(buf,"%s/%s.new",Mailspool,area);
    if((new = fopen(buf,"wt")) == NULLFILE) {
        fclose(txt);
        fclose(idx);
        no_area(m);
		tprintf(Noaccess,buf);
		rmlock(Mailspool,m->area);
		return -1;
    }
    sprintf(buf,"%s/%s.idn",Mailspool,area);
    if((idxnew = fopen(buf,"wt")) == NULLFILE) {
        fclose(txt);
        fclose(idx);
        fclose(new);
        no_area(m);
		tprintf(Noaccess,buf);
		rmlock(Mailspool,m->area);
		return -1;
    }
    /* Copy messages and delete msgs, add status or fwd headers */
    for(cmsg=&m->mbox[1],i = 1; i <= m->nmsgs; i++, cmsg++) {
        /* Read the index line for this message */
        if(fgets(idxline,sizeof(idxline),idx) == NULL) {
            puts("\n\007\007 Warning: Index files not in sync ! Run MAIL2IDX asap!!\n");
            mail_error("Warning: Index files not in sync ! Run MAIL2IDX asap!!\n");
            fclose(idx);
            fclose(txt);
            fclose(new);
            fclose(idxnew);
            no_area(m);
            return -1;
        }
        keep = 0;
        if(!(cmsg->status & BM_DELETE)) {
            /* Keep this message ! */
            keep = 1;
            start = ftell(new);
            fseek(txt,cmsg->start,0);
            /* Copy this one to the new file,
             * start with the smtp header
             * The received: line is first
             */
            while(bgets(buf,sizeof(buf),txt) != NULL) {
                pwait(NULL);
                if(*buf == '\0') {
                    /* End of headers */
                    if(cmsg->status & BM_FORWARDED)
                        fprintf(new,"%s%s\n",Hdrs[XFORWARD],m->name);
                    if(idxline[8] != 'R' && cmsg->status & BM_READ &&
                        (m->areatype == PRIVATE)) {
                        fprintf(new,"%sR\n",Hdrs[STATUS]);
                        idxline[8] = 'R';
                    }
                    fputc('\n',new);
                    break;
                }
                fprintf(new,"%s\n",buf);
            }
            /* Now do the message body */
            pos = ftell(txt);
            while(bgets(buf,sizeof(buf),txt) != NULL) {
                pwait(NULL);
                if(!strncmp(buf,"From ",5)) {
                    /* End of message, rewrite index line */
                    sprintf(len,"%ld",pos-start);
                    sprintf(lens,"%6.6s",len);
                    memcpy(&idxline[40],lens,6);
                    fputs(idxline,idxnew);
                    break;
                }
                fprintf(new,"%s\n",buf);
                pos = ftell(new);
            }
        } /* If not message deleted */
    } /* for all messages */
    /* do we need to keep the last one ? */
    if(keep) {
        /* End of message, rewrite index line */
        sprintf(len,"%ld",pos-start);
        sprintf(lens,"%6.6s",len);
        memcpy(&idxline[40],lens,6);
        fputs(idxline,idxnew);
    }
    size = ftell(new);
    /* If we have no errors, then delete the old mail and index files,
     * and rename the new one.
     */
    if(ferror(new) || ferror(idxnew))
        error = 1;
    else
        error = 0;

    fclose(txt);
    fclose(idx);
    fclose(new);
    fclose(idxnew);

    if(!error) {
        sprintf(buf,"%s/%s.txt",Mailspool,area);
        sprintf(old,"%s/%s.new",Mailspool,area);
        unlink(buf);
        if(size)
            rename(old,buf);
        else
            /* remove a zero length file */
            unlink(old);
        sprintf(buf,"%s/%s.idx",Mailspool,area);
        sprintf(old,"%s/%s.idn",Mailspool,area);
        unlink(buf);
        if(size)
            rename(old,buf);
        else
            /* remove a zero length file */
            unlink(old);
    }
	rmlock(Mailspool,m->area);
    no_area(m);
    free(area);
    if(m->areatype == PRIVATE)
        m->mycnt = fsize(buf) / IDXLINESIZE; /* Update the size of our mailbox */
    return 0;
}

/* Returns 1 if name is a public message Area, 0 otherwise */
int
isarea(name)
char *name;
{
    FILE *fp;
    char *cp, *area;
    char buf[LINELEN];

    if((fp = fopen(Arealist,READ_TEXT)) == NULLFILE)
		return 0;
    area = strdup(name);
    dotformat(area);
    while(fgets(buf,sizeof(buf),fp) != NULLCHAR) {
		/* The first word on each line is all that matters */
        firsttoken(buf);
        dotformat(buf);
        if(stricmp(area,buf) == 0) {    /* found it */
			fclose(fp);
            free(area);
            return 1;
		}
	}
    fclose(fp);
    free(area);
	return 0;
}

static int
lockit(m)
struct mbx *m;
{
	int c, cnt = 0;
    char *area;

    area = strdup(m->area);
    dirformat(area);
    while(mlock(Mailspool,area)) {
        pause(1000L);   /* Wait one second */
		if(++cnt == 10) {
			cnt = 0;
            switch (m->state) {
            case MBX_REVFWD:
            case MBX_FORWARD:
            case MBX_TRYING:
                c='A';
                break;
            default:
                c = tkeywait("Mail file is busy, Abort or Retry ? ",1);
                break;
            }
			if (c == 'A' || c == 'a' || c == EOF) {
                m->nmsgs = 0;
                m->mboxsize = 0;
                free(area);
				return 1;
			}
		}
	}
    free(area);
	return 0;
}

/* read the next message or the current one if new */
int
doreadnext(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
	char buf[10], *newargv[2];

    m = (struct mbx *) p;

    if(m->nmsgs == 0)   /* No mail ! */
        return 0;

    if((m->mbox[m->current].status & BM_READ) != 0) {
		if (m->current == 1 && m->anyread == 0)
			;
		else if (m->current < m->nmsgs) {
			m->current++;
		} else {
            tputs("Last message\n");
			return 0;
		}
	}
	sprintf(buf,"%d",m->current);
    newargv[0] = "r";
	newargv[1] = buf;
	m->anyread = 1;
	return doreadmsg(2,newargv,p);
}

extern int MbRead;

/* display message on the crt given msg number */
/* Modified to allow the 'RM' command, 920307 - WG7J */
int
doreadmsg(argc,argv,p)
int argc;
char *argv[];
void *p;
{
	struct mbx *m;
    int c, col, lin;
    char *area, *cp, *cp2;
    int msg, cnt, i, usemore=0, verbose;
    int smtpheader, mbxheader, pathheader, print;
    FILE *txt;
    char *myargv[NARG];
    int myargc;
    int maxmsg;
    struct let *cmsg;
    char *tmpbuf;
    char buf[LINELEN];

	m = (struct mbx *) p;

    /*Check for read-permissions - WG7J */
    if(m->privs & NO_READCMD) {
        tputs(Noperm);
        return 0;
    }
    if (m->nmsgs == 0) {
        tputs(Nomail);
        return 0;
    }
    if((lin=m->morerows) != 0)
        usemore = 1;    /* Display More prompt */

    /*If this is the RM or VM command, setup myargv[]
     *to contain up to NARG message numbers - WG7J
     */
    if(m->stype == 'M') {
        if(!m->newmsgs) {
            tputs(Nomail);
            return 0;
        }
        /* scan all messsages to find unread ones */
        cmsg = &m->mbox[1];
        myargc = 1;
        i = 0;
        while(myargc < NARG && i < m->nmsgs) {
            if(!(cmsg->status & BM_READ)) { /*found an unread msg!*/
                tmpbuf = mallocw(17); /*allocate space for the new argument*/
                myargv[myargc++] = itoa(i+1,tmpbuf,10);
            }
            i++;
            cmsg++;
        }
        argc = myargc;
    } else {
        /*simply point to the old arguments*/
        for(i=1;i<argc;i++)
            myargv[i] = argv[i];
    }
    if(argc == 1) {
        tputs("Usage: Read/Verbose #\n");
        return 0;
    }
    m->state = MBX_READ;

    /* Open the text file */
    area = strdup(m->area);
    dirformat(area);
    sprintf(buf,"%s/%s.txt",Mailspool,area);
    free(area);
    if((txt=fopen(buf,"rt")) == NULL) {
        tprintf("Cannot open mail file %s\n",area);
        return 0;
    }

    if((*argv[0] == 'v') || (m->stype == 'H'))
        verbose = 1;    /* display all header lines */
    else
        verbose = 0;

    for(i = 1; i < argc; ++i) {
        tmpbuf = strchr(myargv[i],'-');   /* N5KNX: allow from-to msg specification */
        msg = atoi(myargv[i]);
        if (tmpbuf == NULLCHAR)
            maxmsg = msg;
        else
            maxmsg = atoi(++tmpbuf);
        if (maxmsg < msg) {
            tprintf(Badmsg,myargv[i]);
            continue;
        }
        for (; msg <= maxmsg; msg++) {
            if(msg < 1 || msg > m->nmsgs) {
                tprintf(Badmsg,msg);
                goto iamdone;
            }
            cmsg = &m->mbox[msg];
            /* if the message is on hold, only show to sysops - WG7J */
            if((cmsg->status & BM_HOLD) && !(m->privs & SYSOP_CMD)) {
                tprintf("Msg %d:",msg);
                tputs(Noperm);
                continue;
            }
            MbRead++;
#ifdef USERLOG
            /* Check the ID number of this message and
             * adjust new lastread count, if needed - WG7J
             */
            if(cmsg->msgid > m->lastread)
                m->newlastread = cmsg->msgid;
#endif
            m->current = msg;

            /* Only mark your own private area as read and changed.
             * other areas, only mark as read, NOT changed !
             * 910312 - WG7J
             */
            if(!(cmsg->status & BM_READ)) {
                cmsg->status |= BM_READ;
                m->newmsgs--;
                if(m->areatype == PRIVATE)
                    m->change = 1;
            }
            tprintf("Message #%d %s\n", msg,
                cmsg->status & BM_DELETE ? "[Deleted]" : "");
            --lin;

            fseek(txt,cmsg->start,0);
            fgets(buf,sizeof(buf),txt);      /* the 'From ' line */
            if(verbose) {
                tputs(buf);
                lin--;
            }
            smtpheader = 1;
            while(fgets(buf,sizeof(buf),txt) != NULL) {
                if(!strncmp(buf,"From ",5))
                    break;
                print = 1;
                if(!verbose) {
                    if(smtpheader) {
                        if(*buf == '\n') {
                            smtpheader = 0;
                            mbxheader = 1;
                            pathheader = 0;
                        } else {
                            switch(htype(buf)) {
                            case TO:
                            case CC:
                            case FROM:
                            case DATE:
                            case SUBJECT:
                            case REPLYTO:
                            case APPARTO:
                                break;
                            default:
                                print = 0;
                            }
                        }
                    } else if(mbxheader) {
                        if(*buf == '\n') {
                            mbxheader = 0;
                        }
                        print = 0;
                        if(strncmp(buf,"R:",2) || \
                          ((cp=strchr(buf,'@')) == NULL) ) {
                            print = 1;
                            mbxheader = 0;
                            if(pathheader)
                                tputc('\n');
                        } else {
                            if(pathheader > 7)
                                continue;
                            if(*++cp == ':')
                                cp++;
                            cp2 = cp;
                            while(*cp2 != '.' && *cp2 != ' ' && *cp2 != '\n')
                                cp2++;
                            *cp2 = 0;
                            if(++pathheader == 1) {
                                tputs(Hdrs[PATH]);
                                tputs(cp);
                            } else {
                                pathheader++;
                                tputc('!');
                                tputs(cp);
                            }
                        }
                    }
                }
                if(print) {
                    tputs(buf);
                    if(usemore && --lin == 0){
                        if(m->type == TELNET_LINK || m->type == TIP_LINK)
                            c = tkeywait("--More--",0);
                        else  /* For AX.25 and NET/ROM connects - WG7J */
                            c = mykeywait("More(N=no)? ",m);
                        lin = m->morerows;
                        if(c == -1 || c == 'q' || c == 'Q')
                            break;
                        if(c == '\n' || c == '\r')
                            lin = 1;
                    }
                }
            }
        }
	}
iamdone:
    fclose(txt);
    /* If this was 'RM' or 'VM',
     * free the memory allocated for myargv[] - WG7J
     */
    if(m->stype == 'M') {
        for(i=1;i<argc;i++)
            free(myargv[i]);
    }
    return 0;
}

/* Set up m->to when replying to a message. The subject is returned in
 * m->line.
 */
int
mbx_reply(argc,argv,m,cclist,rhdr)
int argc;
char *argv[];
struct mbx *m;
struct list **cclist;	/* Pointer to buffer for pointers to cc recipients */
char **rhdr;		/* Pointer to buffer for extra reply headers */
{
	char subject[MBXLINE], *msgid = NULLCHAR, *date = NULLCHAR;
	char *cp;
	int msg, lastheader, header = NOHEADER;
	long size;

	/* Free anything that might be allocated
	 * since the last call to mbx_to() or mbx_reply()
	 */
	free(m->to);
	m->to = NULLCHAR;
	free(m->tofrom);
	m->tofrom = NULLCHAR;
	free(m->tomsgid);
	m->tomsgid = NULLCHAR;
	free(m->origto);
	m->origto = NULLCHAR;
	subject[0] = '\0';

	if(argc == 1)
	     msg = m->current;
	else
	     msg = atoi(argv[1]);

    /* now open the mailbox file for reading */
    cp = strdup(m->area);
    dirformat(cp);
    sprintf(subject,"%s/%s.txt",Mailspool,cp);
    if ((m->mfile=fopen(subject,"rt")) == NULLFILE) {
        free(cp);
        if(m->sid & MBX_SID)
            tputs("NO - ");
		tputs(Nomail);
        return -1;
	}
    free(cp);
    if(msg < 1 || msg > m->nmsgs) {
        if(m->sid & MBX_SID)
            tputs("NO - ");
        tputs(Badmsg);
        return -1;
	}
	fseek(m->mfile,m->mbox[msg].start,0);
	size = m->mbox[msg].size;
	m->current = msg;
	while(size > 0 && fgets(m->line,MBXLINE-1,m->mfile) != NULLCHAR) {
	     size -= strlen(m->line);
	     if(m->line[0] == '\n')	/* end of header */
		  break;
	     rip(m->line);
	     lastheader = header;
	     if(!isspace(m->line[0])) {
		  header = htype(m->line);
		  lastheader = NOHEADER;
	     }
	     switch(header) {
	     case SUBJECT:
		  if(strlen(m->line) > 11 && !strnicmp(&m->line[9],"Re:",3))
		       strcpy(subject,&m->line[9]);
		  else
		       sprintf(subject,"Re: %s",&m->line[9]);
		  break;
	     case FROM:
		  if(m->to == NULLCHAR && (cp = getaddress(m->line,0)) !=
		     NULLCHAR)
		       m->to = strdup(cp);
		  break;
	     case REPLYTO:
		  if((cp = getaddress(m->line,0)) != NULLCHAR) {
		       free(m->to);
		       m->to = strdup(cp);
		  }
		  break;
	     case MSGID:
		  free(msgid);
		  msgid = strdup(&m->line[12]);
		  break;
	     case DATE:
		  free(date);
		  date = strdup(&m->line[6]);
		  break;
#ifdef notdef
         /* don't want a reply back to myself - WG7J */
         case TO:
         case APPARTO:
#endif
         case CC:
		  /* Get addresses on To, Cc and Apparently-To lines */
		  cp = m->line;
		  m->line[strlen(cp)+1] = '\0';	/* add extra null at end */
		  for(;;) {
		       if((cp = getaddress(cp,lastheader == header ||
					   cp != m->line)) == NULLCHAR)
			    break;
		       addlist(cclist,cp,0);
		       /* skip to next address, if any */
		       cp += strlen(cp) + 1;
		  }
		  break;
	     }
	}
	if(msgid != NULLCHAR || date != NULLCHAR) {
	     *rhdr = mallocw(LINELEN);
	     sprintf(*rhdr,"In-Reply-To: your message ");
	     if(date != NULLCHAR) {
		  sprintf(m->line,"of %s.\n",date);
		  strcat(*rhdr,m->line);
		  if(msgid != NULLCHAR)
		       strcat(*rhdr,"             ");
	     }
	     if(msgid != NULLCHAR) {
		  sprintf(m->line,"%s\n",msgid);
		  strcat(*rhdr,m->line);
	     }
	     free(msgid);
	     free(date);
	}
	strcpy(m->line,subject);
    fclose(m->mfile);
    m->mfile = NULL;
    return 0;
}

#ifdef USERLOG

/*get the last message listed/read
 *from the areaname.USR file
 *keeps track for each user.
 *February '92, WG7J
 */
void
getlastread(m)
struct mbx *m;
{
	FILE *Alog;
	char buf[256];
	char *cp;
	int found=0;

    m->lastread = m->newlastread = 0L;

	sprintf(buf,"%s/%s.usr",Mailspool,m->area);
	if ((Alog = fopen(buf,"r+")) == NULLFILE) {
		/* USR file doesn't exist, create it */
		if((Alog = fopen(buf,"w")) == NULLFILE)
			return;
		/* Add this user as first one */
        sprintf(buf,"%s 0\n",m->name);
		fputs(buf,Alog);
		fclose(Alog);
		return;
	}
	/*Find user in the usr file for this area*/
	for(;;) {
		if(fgets(buf,sizeof(buf),Alog) == NULLCHAR)
            break;
		if((cp=strchr(buf,' ')) != NULLCHAR)
			*cp = '\0';
		if(!stricmp(m->name,buf)) {
			/*found user*/
            cp++;
			while(*cp == ' ')   /*skip blanks*/
				cp++;
			m->lastread = atol(cp);
			found = 1;
			break;
		}
	}
	if(!found) {
		/*Add user*/
        sprintf(buf,"%s 0\n",m->name);
		fputs(buf,Alog);
	}
	fclose(Alog);
	return;
}

/* Write the new last read id number to the USR file - WG7J
 * only update if this is not a bbs,
 * current area is a public area and not 'help',
 * or anything that starts with 'sys',
 * and a new message was actually listed/read
 */
void
setlastread(m)
struct mbx *m;
{
	FILE *Alog, *tfile;
    char *cp;
    char buf[80];
    char tmpname[80];

    if((m->sid & MBX_SID) || (m->newlastread <= m->lastread) )
        return;

    if((isarea(m->area) && strcmp(m->area,"help")) ||
        !strncmp(m->area,"sys",3) ) {

        /* Rename the USR file to a tempfile */
        sprintf(buf,"%s/%s.usr",Mailspool,m->area);

        /* N5KNX: change suffix from .usr to .tmp */
        sprintf(tmpname,"%s/%s.tmp",Mailspool,m->area);
        if(rename(buf,tmpname))
            /* Can't rename ??? */
            return;

        if((Alog = fopen(buf,"w")) == NULLFILE) {
            /* can't creat new USR file ???*/
            rename(tmpname,buf);    /* try to undo the damage */
            return;
        }

        if((tfile = fopen(tmpname,"r")) == NULLFILE)
            /* can't open renamed file ??? */
            return;

        /*Write all users back, but update this one!*/
        while(fgets(buf,sizeof(buf),tfile) != NULLCHAR) {
            if((cp=strchr(buf,' ')) != NULLCHAR)
                *cp = '\0';
            if(!stricmp(m->name,buf)) {
                /*found this user*/
                sprintf(buf,"%s %lu\n",m->name,m->newlastread);
            } else
                *cp = ' '; /* restore the space !*/
            fputs(buf,Alog);
        }
        fclose(tfile);
        unlink(tmpname);
        fclose(Alog);
    }
    return;
}

#endif /*USERLOG*/

void
scanmail(m)		 /* Get any new mail */
struct mbx *m;
{
    FILE *idx;
    char *area;
    int ret;
    char buf[128];
	long diff;

    if ((diff = isnewmail(m)) == 0L)
		return;
	if(lockit(m))
        return;
    if(diff < 0L) {
        /* Area wasn't open yet, or the file size has decreased.
         * Any changes we did to this area will be lost,
         * but this is not fatal.
		 */
        initnotes(m);
		rmlock(Mailspool,m->area);
		return;
	}
    area = strdup(m->area);
    dirformat(area);
    sprintf(buf,"%s/%s.idx",Mailspool,area);
    if ((idx = fopen(buf,"r")) == NULLFILE)
        tprintf(Noaccess,buf);
	else {
        /* Reread all messages since size they may have changed
		 * in size after a X-Forwarded-To line was added.
		 */
        ret = readnotes(m,idx);   /* get the mail */
        (void) fclose(idx);
        if(m->areatype == PRIVATE)
            m->mycnt = m->nmsgs;
		if (ret != 0)
            tputs("Error updating mail file\n");
	}
    rmlock(Mailspool,area);
    free(area);
}

/* Check the current mailbox to see if new mail has arrived.
 * Returns the difference in size.
 */
static long
isnewmail(m)
struct mbx *m;
{
    long size;
    char buf[128];

    sprintf(buf,"%s/%s",Mailspool,m->area);
    dirformat(buf);
    strcat(buf,".idx");
    if((size=fsize(buf)) <= 0) {
        if(m->mboxsize == 0)
            return 0;
        return -1;
    }
    /* index file exists, check mailbox */
    strcpy(&buf[strlen(buf) - 3],"txt");
    size = fsize(buf);
    if(m->mboxsize == 0) {
        if(size > 0)
            return -1;  /* New open of mailbox */
        return 0;       /* No mail; no change */
    }
    /* Others */
    return size - m->mboxsize;
}

/* Check if the private mail area has changed */
long
isnewprivmail(m)
struct mbx *m;
{
    int msgcnt;
    long size;
	char buf[256];

    sprintf(buf,"%s/%s.idx",Mailspool,m->name);
    if((size = fsize(buf)) <= 0)
        msgcnt = 0;
    else
        msgcnt = size / IDXLINESIZE;
    return msgcnt - m->mycnt; /* != 0 not more than once */
}

#endif /* MAILCMDS */

/* This function returns the length of a file. The proper thing would be
 * to use stat(), but it fails when using DesqView together with Turbo-C
 * code.
 */
long
fsize(name)
char *name;
{
	long cnt;
	FILE *fp;

    if((fp = fopen(name,READ_TEXT)) == NULLFILE)
		return -1L;
	fseek(fp,0L,2);
	cnt = ftell(fp);
	fclose(fp);
	return cnt;
}

/* Print prompt and read one character, telnet version */
int
tkeywait(prompt,flush)
char *prompt;	/* Optional prompt */
int flush;	/* Flush queued input? */
{
	int c, i, oldimode,oldomode;

	if(flush && socklen(Curproc->input,0) != 0)
		recv_mbuf(Curproc->input,NULL,0,NULLCHAR,0); /* flush */
	if(prompt == NULLCHAR)
		prompt = "Hit enter to continue"; 
	tprintf("%s%c%c%c",prompt,IAC,WILL,TN_ECHO);
	usflush(Curproc->output);

	/* discard the response */

	oldimode = sockmode(Curproc->input,SOCK_BINARY);
	oldomode = sockmode(Curproc->output,SOCK_BINARY);

	while((c = rrecvchar(Curproc->input)) == IAC){
		c = rrecvchar(Curproc->input);
		if(c > 250 && c < 255)
			rrecvchar(Curproc->input);
	}

	sockmode(Curproc->output,oldomode);
	sockmode(Curproc->input,oldimode);

	/* Get rid of the prompt */
	for(i=strlen(prompt);i != 0;i--)
		tputc('\b');
	for(i=strlen(prompt);i != 0;i--)
		tputc(' ');
	for(i=strlen(prompt);i != 0;i--)
		tputc('\b');
	tprintf("%c%c%c",IAC,WONT,TN_ECHO);
	usflush(Curproc->output);
	return c;
}

/* Print prompt and read reply,
 * AX.25 and NETROM version - WG7J
 * 'N' or 'n' returns -1, everything else return 0.
 */
int
mykeywait(prompt,m)
char *prompt;
struct mbx *m;
{
    tputs(prompt);
    usflush(Curproc->output);
    if(recvline(m->user, m->line, MBXLINE) == -1) {
        return -1;
    }
    /* Only 'N' or 'n' really matters */
    if((*m->line == 'N') || (*m->line == 'n'))
        return -1;
    return 0;
}

void dotformat(char *area) {

    while(*area) {
        if(*area == '\\' || *area == '/')
            *area = '.';
        area++;
    }
}

void dirformat(char *area) {

    while(*area) {
        if(*area == '\\' || *area == '.')
            *area = '/';
        area++;
    }
}

#ifdef MAILCMDS

void firsttoken(char *line) {

    while(*line) {
        if(*line == ' ' || *line == '\n' || *line == '\t') {
            *line = '\0';
            break;
        }
        line++;
    }
}
#endif /* MAILCMDS */

/* Get a line from the text file that was opened in binary mode.
 * Delete the CR/LF at the end !
 */
char *bgets(char * buf,int size,FILE *fp) {
    char *cp;

    if(fgets(buf,size,fp) == NULL)
        return NULL;
    cp = buf;
    while(*cp) {
        if(*cp == 0xd || *cp == 0xa)
            *cp = '\0';
        cp++;
    }
    return buf;
}

#ifdef notdef

/* Echo some stuff for debugging */
void eout(char *s) {

    puts(s);
    fflush(stdout);
    while(!kbhit());
    getch();
}

#endif

