/*	General Mail Utilities
 *	culled from other files to reduce unnecessary cross references.
 *	This material is believed to be in the public domain.
 */
#ifdef MSDOS
#include <io.h>
#endif
#include <fcntl.h>
#include <ctype.h>
#include "global.h"
#include "socket.h"
#include "mailutil.h"
#include "mailbox.h"
#include "smtp.h"
#include "files.h"
#include "bm.h"

#ifdef linux
extern int close __ARGS((int)); /* this should be in fcntl.h... */
extern int unlink __ARGS((char *)); /* can't include unistd.h, collisions */
#endif

/*  Jan 92  Bill Simpson
 *		The following routines were combined from smtpserv.c,
 *		pop3cli.c and nntpcli.c
 */

/* June 93 - Johan. K. Reinalda, WG7J
 * Mail file indexing added.
 */

#ifdef POP3CLIENT

/* Receive message from socket, copying to file.
 * Returns number of lines received, -1 indicates error
 */
int
recvmail(s,buf,len,fp,trace)
int s;		/* Socket index */
char *buf;	/* Line buffer */
unsigned len;	/* Length of buffer */
FILE *fp;	/* File to copy into */
int trace;	/* to trace or not to trace */
{
	int lines = 0;

	while (recvline(s,buf,len) != -1) {
		register char *p = buf;

		if (trace >= 4)
			log(s,"<==%s", buf);

		/* check for end of message . or escaped .. */
		if (*p == '.') {
			if (*++p == '\n') {
				if (trace >= 3)
					log(s,"received %d lines", lines);
				return lines;
			} else if ( *p != '.' ) {
				p--;
			}
		} else if (strncmp(p,"From ",5) == 0) {
			/* for UNIX mail compatiblity */
			(void) putc('>',fp);
		}
		/* Append to data file */
		fputs(p,fp);
		++lines;
	}
	if ( trace >= 1 )
		log(s,"receive error after %d lines", lines);
	return -1;
}
#endif

#if (defined(POP2CLIENT) || defined(POP3CLIENT))

/* Copy from the work file into the mailbox.
 * -1 indicates error
 */
int
copymail(filename,buf,len,wfp,trace)
char *filename; /* Target Filename */
char *buf;	/* Line buffer */
unsigned len;	/* Length of buffer */
FILE *wfp;	/* File to copy from */
int trace;      /* to trace or not to trace */
{
    FILE *idx, *mfp = NULLFILE;
    int prev,header;
    long pos,start;
    char *cp;
    char l[10];
    struct mailindex ind;

    while (mlock(Mailspool,filename)) {
		pause(10000L);	/* 10 seconds */
	}

    sprintf(buf,"%s/%s.txt",Mailspool,filename);
    if ((mfp = fopen(buf,APPEND_TEXT)) == NULL) {
		if ( trace >= 1 )
			log(-1,"*** Unable to open %s!!!\n", buf);
		tprintf("\n*** Unable to open %s!!!\n", buf);
        rmlock(Mailspool,filename);
		return -1;
	}

    rewind(wfp);            /* start of new mail, just to be safe ! */
    fseek(mfp,0,SEEK_END);  /* End of mailbox file */
    prev = 0;
    pos = ftell(mfp);

    /* loop for all lines in file */
    while(fgets(buf,len,wfp) != NULL) {
        if(!strncmp(buf,"From ",5)) {
            /* start of next message */
            if(prev) {
                /* Now add the index line of the previous message */
                ind.size = ftell(mfp) - start;
                write_index(filename,&ind);
            }
            default_index(filename,&ind);
            start = pos;
            prev = 1;
            header = 1;
            /* Read the 'Received...' and 'ID... lines'
             * to get the msgid - WG7J
             */
            fputs(buf, mfp);        /* First write the 'From ' line */
            fgets(buf, len, wfp);   /* "Received ' line */
            fputs(buf, mfp);
            fgets(buf, len, wfp);   /* 'id' line */
            fputs(buf, mfp);
            if((cp=strstr(buf,"AA")) != NULLCHAR)
                /* what follows is the message-number */
                ind.msgid = atol(cp+2);
        } else {
            /* Now the rest of the message */
            fputs(buf,mfp); /* Write to the mailbox */
            if(*buf == '\n')
                header = 0;
            if(header)
                set_index(buf,&ind);
            pwait(NULL);    /* give other processes time in long copy */
        }
        pos = ftell(mfp);
    }
    if(prev) {
        /* Now add the index line of the previous message */
        ind.size = ftell(mfp) - start;
        write_index(filename,&ind);
    }
    fclose(mfp);
    rmlock(Mailspool,filename);
	return 0;
}
#endif


/* 	Jan 92	Bill Simpson
 * 		The following routines were extracted from smtpserv.c
 *		and smtpcli.c, since they are used from several places.
 */

/* create mail lockfile */
int
mlock(dir,id)
char *dir,*id;
{
    int fd;
    char lockname[FILE_PATH_SIZE];

	/* Try to create the lock file in an atomic operation */
    sprintf(lockname,"%s/%s",dir,id);
    dirformat(lockname);
    strcat(lockname,".lck");

#ifdef        AMIGA
	/* don't ask, really, just don't ask... I'd do file locking on
	 * an Amiga much more differently than this.
	 */
	if(access(lockname, 0) == 0)
		return -1;
#endif
	if((fd = open(lockname, O_WRONLY|O_EXCL|O_CREAT,0600)) == -1)
		return -1;
	close(fd);
	return 0;
}


/* remove mail lockfile */
int
rmlock(dir,id)
char *dir,*id;
{
	char lockname[FILE_PATH_SIZE];

    sprintf(lockname,"%s/%s",dir,id);
    dirformat(lockname);
    strcat(lockname,".lck");
	return(unlink(lockname));
}


/* Given a string of the form <user@host>, extract the part inside the
 * angle brackets and return a pointer to it.
 */
char *
getname(cp)
register char *cp;
{
	register char *cp1;

	if ((cp = strchr(cp,'<')) == NULLCHAR)
		return NULLCHAR;
	cp++;	/* cp -> first char of name */
	if ((cp1 = strchr(cp,'>')) == NULLCHAR)
		return NULLCHAR;
	*cp1 = '\0';
	return cp;
}


/*      Jan 92  Bill Simpson
 *		The following routines were extracted from mailbox.c,
 *		since they are used from several places.
 */


/* Parse a string in the "Text: Text <user@host>" or "Text: user@host (Text)"
 * formats for the address user@host.
 */
char *
getaddress(string,cont)
char *string;
int cont;		/* true if string is a continued header line */
{
	char *cp, *ap = NULLCHAR;
	int par = 0;

	if((cp = getname(string)) != NULLCHAR) /* Look for <> style address */
		return cp;

	cp = string;
	if(!cont)
		if((cp = strchr(string,':')) == NULLCHAR)	/* Skip the token */
			return NULLCHAR;
		else
			++cp;
	for(; *cp != '\0'; ++cp) {
		if(par && *cp == ')') {
			--par;
			continue;
		}
		if(*cp == '(')		/* Ignore text within parenthesis */
			++par;
		if(par)
			continue;
		if(*cp == ' ' || *cp == '\t' || *cp == '\n' || *cp == ',') {
			if(ap != NULLCHAR)
				break;
			continue;
		}
		if(ap == NULLCHAR)
			ap = cp;
	}
	*cp = '\0';
	return ap;
}


char *Hdrs[] = {
	"Approved: ",
	"From: ",
	"To: ",
	"Date: ",
	"Message-Id: ",
	"Subject: ",
	"Received: ",
	"Sender: ",
	"Reply-To: ",
	"Status: ",
	"X-BBS-Msg-Type: ",
	"X-Forwarded-To: ",
	"Cc: ",
	"Return-Receipt-To: ",
	"Apparently-To: ",
	"Errors-To: ",
	"Organization: ",
    "Newsgroup: ",
    "Path: ",
    "X-BBS-Hold: ",
    NULLCHAR,
};


/* return the header type */
int
htype(s)
char *s;
{
	register char *p;
	register int i;

	p = s;
	/* check to see if there is a ':' before and white space */
	while (*p != '\0' && *p != ' ' && *p != ':')
		p++;
	if (*p != ':')
		return NOHEADER;

	for (i = 0; Hdrs[i] != NULLCHAR; i++) {
		if (strnicmp(Hdrs[i],s,strlen(Hdrs[i])) == 0)
			return i;
	}
	return UNKNOWN;
}


/* Read the rewrite file for lines where the first word is a regular
 * expression and the second word are rewriting rules. The special
 * character '$' followed by a digit denotes the string that matched
 * a '*' character. The '*' characters are numbered from 1 to 9.
 * Example: the line "*@*.* $2@$1.ampr.org" would rewrite the address
 * "foo@bar.xxx" to "bar@foo.ampr.org".
 * $H is replaced by our hostname, and $$ is an escaped $ character.
 * If the third word on the line has an 'r' character in it, the function
 * will recurse with the new address.
 */
char *
rewrite_address(addr)
char *addr;
{
	char *argv[10], buf[PLINELEN], *cp, *cp2, *retstr;
	int cnt;
	FILE *fp;

	if ((fp = fopen(Rewritefile,READ_TEXT)) == NULLFILE)
		return NULLCHAR;

	memset((char *)argv,0,10*sizeof(char *));
	while(fgets(buf,sizeof(buf),fp) != NULLCHAR) {
		if(*buf == '#')		/* skip commented lines */
			continue;

        if((cp = strchr(buf,' ')) == NULLCHAR) /* get the first word */
			if((cp = strchr(buf,'\t')) == NULLCHAR)
				continue;
		*cp = '\0';
		if((cp2 = strchr(buf,'\t')) != NULLCHAR){
			*cp = ' ';
			cp = cp2;
			*cp = '\0';
		}
		if(!wildmat(addr,buf,argv))
			continue;		/* no match */
		rip(++cp);
        /* scan past additional whitespaces */
        while (*cp == ' ' || *cp == '\t') ++cp;
        cp2 = retstr = (char *) callocw(1,PLINELEN);
		while(*cp != '\0' && *cp != ' ' && *cp != '\t')
			if(*cp == '$') {
				if(isdigit(*(++cp)))
					if(argv[*cp - '0'-1] != '\0')
						strcat(cp2,argv[*cp - '0'-1]);
				if(*cp == 'h' || *cp == 'H') /* Our hostname */
					strcat(cp2,Hostname);
				if(*cp == '$')	/* Escaped $ character */
					strcat(cp2,"$");
				cp2 = retstr + strlen(retstr);
				cp++;
			}
			else
				*cp2++ = *cp++;
		for(cnt=0; argv[cnt] != NULLCHAR; ++cnt)
			free(argv[cnt]);
		fclose(fp);
		/* If there remains an 'r' character on the line, repeat
		 * everything by recursing.
		 */
		if(strchr(cp,'r') != NULLCHAR || strchr(cp,'R') != NULLCHAR) {
			if((cp2 = rewrite_address(retstr)) != NULLCHAR) {
				free(retstr);
				return cp2;
			}
		}
		return retstr;
	}
	fclose(fp);
	return NULLCHAR;
}

#ifdef notdef

/* Read a file containing a message,
 * and build the index from the smtp header.
 */
void create_index(FILE *fp, char *name, struct mailindex *ind) {
    int header;
    long start,begin,pos;
    char *cp;
    char buf[128];

    default_index(name,ind);

    /* Find beginning of message */
    start = begin = ftell(fp);
    while(fgets(buf, sizeof(buf), fp) != NULLCHAR) {
        /* search 'From ' line */
        if(!strncmp(buf,"From ",5))
            break;
        start = ftell(fp);
    }
    /* Read the 'Received...' and 'ID... lines'
     * to get the msgid - WG7J
     */
    fgets(buf, sizeof(buf), fp);    /* "Receiced " line */
    fgets(buf, sizeof(buf), fp);    /* 'id' line */
    if((cp=strstr(buf,"AA")) != NULLCHAR)
        /* what follows is the message-number */
        ind->msgid = atol(cp+2);
    header = 1;
    pos = ftell(fp);
    while(fgets(buf,sizeof(buf),fp) != NULLCHAR ) {
        if(!strncmp(buf,"From ",5)) /* next message */
            break;
        if(*buf == '\n')
            header = 0;
        rip(buf);
        if(header)
            set_index(buf,ind);
        pos = ftell(fp);
    }
    ind->size = pos - start;
    fseek(fp,begin,SEEK_SET);   /* return to old position in file */
}
#endif

void set_index(char *buf,struct mailindex *index) {
    char *s,*cp;

    switch(htype(buf)) {
    case APPARTO:
        if(*index->to != '\0')
            break;
        /* Notice fall-through */
    case TO:
        cp = getaddress(buf,0);
        sprintf(index->to,"%.13s",cp != NULLCHAR ? cp : "");
        break;
    case STATUS:
        index->status = buf[8];
        break;
    case FROM:
        cp = getaddress(buf,0);
        sprintf(index->from,"%.8s",
            cp != NULLCHAR ? cp : "");
        if((cp=strchr(index->from,'@')) != NULLCHAR)
            *cp = '\0';         /* get rid of @-host or @-bbs field */
        break;
    case SUBJECT:
        cp = &buf[9];
        while(*cp == ' ')
            cp++;
        sprintf(index->subject,"%.35s",cp);
        /* translate possible left over cr/lf's to spaces */
        cp = index->subject;
        while(*cp) {
            if(*cp == '\n' || *cp == 0xd)
                *cp = ' ';
            cp++;
        }
        break;
    case BBSTYPE:
        index->type = buf[16];
        break;
    case DATE:
        /* find age from ARPA style date */
        s = &buf[6];
        while(*s == ' ')
            s++;
        /* check to see if there is a "Day, " field */
        if((cp=strchr(s,',')) != NULLCHAR)  {
            /* probably standard ARPA style header */
            cp += 2; /* get past header and DAY field */
        } else {
            /* probably a NNTP style message, that has no
             * "Day, " part in the date line
             */
             cp = s;
        }
        while(*cp == ' ')
            cp++;
        /* now we should be at the start of the
         * "14 Apr 92 08:14:32" string
         */
        if(strlen(cp) < 17)
            break;
        if(cp[5] == ' ') {
            index->date[0] = ' ';
            strncpy(&index->date[1],cp,5);
        } else {
            strncpy(index->date,cp,6);
            if(index->date[0] == '0')
                index->date[0] = ' ';
        }
        index->date[6] = '\0';
        break;
    } /* switch */
}


/* Update the index file */
void write_index(char *name, struct mailindex *ind) {
    FILE *idx;
    char len[10];
    char buf[128];


    sprintf(buf,"%s/%s.idx",Mailspool,name);
    if((idx=fopen(buf,APPEND_TEXT)) != NULLFILE) {
        sprintf(len,"%ld",ind->size);
        fprintf(idx,"%5.5ld %c %c %-13.13s %-8.8s %-6.6s %6.6s %-35.35s\n",
                    ind->msgid,ind->type,ind->status,ind->to,
                    ind->from,ind->date,len,ind->subject);
        fclose(idx);
    }
}

void default_index(char *name, struct mailindex *ind) {

    *ind->date = '\0';
    *ind->from = '\0';
    strcpy(ind->subject,"(none)");
    *ind->to = '\0';
    ind->status = 'N';
    ind->size = 0;
    ind->msgid = 0;
#ifdef MAILCMDS
#ifdef MAILBOX
    if(isarea(name))
        ind->type = 'B';
    else
#endif
#endif
        ind->type = 'P';
}

void delete_index(char *filename) {
    char idxfile[128];

    sprintf(idxfile,"%s/%s.idx",Mailspool,filename);
    unlink(idxfile);
}

